import sys
import os
from datetime import datetime, timedelta
import logging; logging.captureWarnings(True); logger = logging.getLogger(__name__)
import string
from shutil import move

import numpy as np
from netCDF4 import Dataset, num2date, date2num

# Module with functions to export/import trajectory data to/from netCDF file
# Strives to be compliant with netCDF CF-convention on trajectories
# https://cfconventions.org/Data/cf-conventions/cf-conventions-1.6/build/cf-conventions.html#idp8377728
# https://geo-ide.noaa.gov/wiki/index.php?title=NODC_NetCDF_Trajectory_Template

skip_parameters = ['ID']  # Do not write to file


def init(self, filename):

    self.outfile_name = filename
    self.outfile = Dataset(filename, 'w')
    self.outfile.createDimension('trajectory', self.num_elements_total())
    self.outfile.createVariable('trajectory', 'i4', ('trajectory',))
    self.outfile.createDimension('time', None)  # Unlimited time dimension
    self.outfile.createVariable('time', 'f8', ('time',))
    # NB: trajectory_id must be changed for future ragged array representation
    self.outfile.variables['trajectory'][:] = \
        np.arange(self.num_elements_total())+1
    self.outfile.variables['trajectory'].cf_role = 'trajectory_id'
    self.outfile.variables['trajectory'].units = '1'

    self.outfile.Conventions = 'CF-1.6'
    self.outfile.standard_name_vocabulary = 'CF-1.6'
    self.outfile.featureType = 'trajectory'
    self.outfile.history = 'Created ' + str(datetime.now())
    self.outfile.source = 'Output from simulation with OpenDrift'
    self.outfile.model_url = 'https://github.com/OpenDrift/opendrift'
    self.outfile.opendrift_class = self.__class__.__name__
    self.outfile.opendrift_module = self.__class__.__module__
    self.outfile.readers = str(self.readers.keys())
    self.outfile.time_coverage_start = str(self.start_time)
    self.outfile.time_step_calculation = str(self.time_step)
    self.outfile.time_step_output = str(self.time_step_output)

    # Time
    self.timeStr = 'seconds since 1970-01-01 00:00:00'
    self.outfile.variables['time'].units = self.timeStr
    self.outfile.variables['time'].standard_name = 'time'
    self.outfile.variables['time'].long_name = 'time'
    # Apparently axis attribute shall not be given for time, lon and lat
    #self.outfile.variables['time'].axis = 'T'

    # Write config settings
    for key in self._config:
        value = self.get_config(key)
        if isinstance(value, (bool, type(None))):
            value = str(value)
        self.outfile.setncattr('config_' + key, value)

    # Write additionaly metadata attributes, if given
    if hasattr(self, 'metadata_dict'):
        for key, value in self.metadata_dict.items():
            self.outfile.setncattr(key, str(value))

    # Add all element properties as variables
    for prop in self.history.dtype.fields:
        if prop in skip_parameters:
            continue
        # Note: Should use 'f8' if 'f4' is not accurate enough,
        #       at expense of larger files
        try:
            dtype = self.history.dtype[prop]
        except:
            dtype = 'f4'
        var = self.outfile.createVariable(prop, dtype, ('trajectory', 'time'))
        for subprop in self.history_metadata[prop].items():
            if subprop[0] not in ['dtype', 'constant', 'default', 'seed']:
                # Apparently axis attribute shall not be given for lon and lat:
                if prop in ['lon', 'lat'] and subprop[0] == 'axis':
                    continue
                var.setncattr(subprop[0], subprop[1])

def write_buffer(self):
    if self.outfile._isopen == 0:
        self.outfile = Dataset(self.outfile_name, 'a')
    num_steps_to_export = self.steps_output - self.steps_exported
    for prop in self.history_metadata:
        if prop in skip_parameters:
            continue
        var = self.outfile.variables[prop]
        var[:, self.steps_exported:self.steps_exported+num_steps_to_export] = \
            self.history[prop][:, 0:num_steps_to_export]

    times = [self.start_time + n*self.time_step_output for n in
             range(self.steps_exported, self.steps_output)]
    self.outfile.variables['time'][
        self.steps_exported:self.steps_exported+len(times)] = \
            date2num(times, self.timeStr)

    # Write status categories metadata
    # TODO: need not be written each output timestep, thus this could be deleted?
    #status_dtype = self.ElementType.variables['status']['dtype']
    #self.outfile.variables['status'].valid_range = np.array(
    #    (0, len(self.status_categories) - 1)).astype(status_dtype)
    #self.outfile.variables['status'].flag_values = \
    #    np.array(np.arange(len(self.status_categories)), dtype=status_dtype)
    #self.outfile.variables['status'].flag_meanings = \
    #    " ".join(self.status_categories)

    logger.info('Wrote %s steps to file %s' % (num_steps_to_export,
                                                self.outfile_name))
    self.history.mask = True  # Reset history array, for new data
    self.steps_exported = self.steps_exported + num_steps_to_export
    self.outfile.steps_exported = self.steps_exported
    self.outfile.sync()  # Flush from memory to disk
    self.outfile.close()  # close file temporarily

def close(self):
    self.outfile = Dataset(self.outfile_name, 'a')
    # Write status categories metadata
    status_dtype = self.ElementType.variables['status']['dtype']
    self.outfile.variables['status'].valid_range = np.array(
        (0, len(self.status_categories) - 1)).astype(status_dtype)
    self.outfile.variables['status'].flag_values = \
        np.array(np.arange(len(self.status_categories)), dtype=status_dtype)
    self.outfile.variables['status'].flag_meanings = \
        " ".join(self.status_categories)

    # Write origin_marker definitions
    if 'origin_marker' in self.outfile.variables:
        self.outfile.variables['origin_marker'].flag_values = \
            np.array(np.arange(len(self.origin_marker)))
        self.outfile.variables['origin_marker'].flag_meanings = \
            " ".join(self.origin_marker.values())

    # Write final timesteps to file
    self.outfile.time_coverage_end = str(self.time)

    # Write performance data
    self.outfile.performance = self.performance()

    # Write metadata items anew, if any are added during simulation
    if hasattr(self, 'metadata_dict'):
        for key, value in self.metadata_dict.items():
            self.outfile.setncattr(key, str(value))

    # Write min and max values as variable attributes
    for var in self.history_metadata:
        if var in self.outfile.variables:
            self.outfile.variables[var].setncattr('minval', self.minvals[var])
            self.outfile.variables[var].setncattr('maxval', self.maxvals[var])

    # Write bounds metadata
    self.outfile.geospatial_lat_min = self.history['lat'].min()
    self.outfile.geospatial_lat_max = self.history['lat'].max()
    self.outfile.geospatial_lat_units = 'degrees_north'
    self.outfile.geospatial_lat_resolution = 'point'
    self.outfile.geospatial_lon_min = self.history['lon'].min()
    self.outfile.geospatial_lon_max = self.history['lon'].max()
    self.outfile.geospatial_lon_units = 'degrees_east'
    self.outfile.geospatial_lon_resolution = 'point'
    self.outfile.runtime = str(datetime.now() -
                               self.timers['total time'])

    self.outfile.close()  # Finally close file

    # Finally changing UNLIMITED time dimension to fixed, for CDM compliance.
    # Fortunately this is quite fast.
    # https://www.unidata.ucar.edu/software/thredds/current/netcdf-java/reference/FeatureDatasets/CFpointImplement.html
    try:
        logger.debug('Making netCDF file CDM compliant with fixed dimensions')
        if self.num_elements_scheduled() > 0:
            logger.info('Removing %i unseeded elements already written to file' % self.num_elements_scheduled())
            mask = np.ones(self.history.shape[0], dtype=bool)
            mask[self.elements_scheduled.ID-1] = False
        with Dataset(self.outfile_name) as src, \
                Dataset(self.outfile_name + '_tmp', 'w') as dst:
            for name, dimension in src.dimensions.items():
                if name=='trajectory':
                    # Truncate dimension length to  number actually seeded
                    dst.createDimension(name, self.num_elements_activated())
                else:
                    dst.createDimension(name, len(dimension))

            for name, variable in src.variables.items():
                dstVar = dst.createVariable(name, variable.datatype,
                                             variable.dimensions)
                srcVar = src.variables[name]
                # Truncate data to number actually seeded
                if 'trajectory' in variable.dimensions:
                    if self.num_elements_scheduled() > 0:
                        if len(variable.dimensions) == 2:
                            dstVar[:] = srcVar[mask, :]
                        else:
                            dstVar[:] = srcVar[mask]  # Copy data
                    else:
                        dstVar[:] = srcVar[:]
                else:
                    dstVar[:] = srcVar[:]
                for att in src.variables[name].ncattrs():
                    # Copy variable attributes
                    dstVar.setncattr(att, srcVar.getncattr(att))

            for att in src.ncattrs():  # Copy global attributes
                dst.setncattr(att, src.getncattr(att))

        move(self.outfile_name + '_tmp', self.outfile_name)  # Replace original
    except Exception as me:
        print(me)
        print('Could not convert netCDF file from unlimited to fixed dimension. Could be due to netCDF library incompatibility(?)')

def import_file_xarray(self, filename, chunks):

    import xarray as xr
    logger.debug('Importing with Xarray from ' + filename)
    self.ds = xr.open_dataset(filename, chunks=chunks)

    self.steps_output = len(self.ds.time)
    ts0 = (self.ds.time[0] - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's')
    self.start_time = datetime.utcfromtimestamp(float(ts0))
    tse = (self.ds.time[-1] - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's')
    self.end_time = datetime.utcfromtimestamp(float(tse))
    if len(self.ds.time) > 1:
        ts1 = (self.ds.time[1] - np.datetime64('1970-01-01T00:00:00')) / np.timedelta64(1, 's')
        self.time_step_output = timedelta(seconds=float(ts1 - ts0))
    self.time = self.end_time  # Using end time as default
    self.status_categories = self.ds.status.flag_meanings.split()
    if 'flag_meanings' in self.ds.origin_marker.attrs:
        self.origin_marker = [s.replace('_', ' ') for s in self.ds.origin_marker.flag_meanings.split()]

    num_elements = len(self.ds.trajectory)
    elements=np.arange(num_elements)

    # Data types for variables
    dtype = np.dtype([(var[0], var[1]['dtype'])
                      for var in self.ElementType.variables.items()])
    history_dtype_fields = [
        (name, self.ElementType.variables[name]['dtype'])
        for name in self.ElementType.variables]
    # Add environment variables
    self.history_metadata = self.ElementType.variables.copy()
    for env_var in self.required_variables:
        if env_var in self.ds.variables:
            history_dtype_fields.append((env_var, np.dtype('float32')))
            self.history_metadata[env_var] = {}
    history_dtype = np.dtype(history_dtype_fields)

    # Masking where elements are not been seeded
    # TODO: mask from deactivation towards end
    for da in ['lon', 'lat']:
        self.ds[da] = self.ds[da].where(self.ds.status>=0)

    if 'minval' in self.ds.lon.attrs:
        self.lonmin = np.float32(self.ds.lon.minval)
        self.latmin = np.float32(self.ds.lat.minval)
        self.lonmax = np.float32(self.ds.lon.maxval)
        self.latmax = np.float32(self.ds.lat.maxval)

def import_file(self, filename, times=None, elements=None, load_history=True):
    """Create OpenDrift object from imported file.
     times: indices of time steps to be imported, must be contineous range.
     elements: indices of elements to be imported
    """

    logger.debug('Importing from ' + filename)
    infile = Dataset(filename, 'r')
    # 'times' can be used to import subset. Not yet implemented.
    if times is None and hasattr(infile, 'steps_exported'):
        self.steps_output = infile.steps_exported
        times = np.arange(infile.steps_exported)
    else:
        #self.steps_output = len(infile.dimensions['time'])
        self.steps_output = len(times)

    filetime = infile.variables['time'][times]
    units = infile.variables['time'].units
    self.start_time = num2date(filetime[0], units)
    if len(filetime) > 1:
        self.end_time = num2date(filetime[self.steps_output-1], units)  # Why -1?
        self.time_step_output = num2date(filetime[1], units) - self.start_time
    else:
        self.time_step_output = timedelta(hours=1)
        self.end_time = self.start_time
    self.time = self.end_time  # Using end time as default
    self.status_categories = infile.variables['status'].flag_meanings.split()

    if elements is None:
        num_elements = len(infile.dimensions['trajectory'])
        elements=np.arange(num_elements)
    else:
        num_elements=len(elements)

    dtype = np.dtype([(var[0], var[1]['dtype'])
                      for var in self.ElementType.variables.items()])

    history_dtype_fields = [
        (name, self.ElementType.variables[name]['dtype'])
        for name in self.ElementType.variables]
    # Add environment variables
    self.history_metadata = self.ElementType.variables.copy()
    for env_var in self.required_variables:
        if env_var in infile.variables:
            history_dtype_fields.append((env_var, np.dtype('float32')))
            self.history_metadata[env_var] = {}
    history_dtype = np.dtype(history_dtype_fields)

    # Import dataset (history)
    if load_history is True:
        self.history = np.ma.array(
            np.zeros([num_elements, self.steps_output]),
            dtype=history_dtype, mask=[True])
        for var in infile.variables:
            if var in ['time', 'trajectory']:
                continue
            try:
                self.history[var] = infile.variables[var][elements, times]
            except Exception as e:
                logger.info(e)
                pass

        # Initialise elements from given (or last) state/time
        firstlast = np.ma.notmasked_edges(self.history['status'], axis=1)
        index_of_last = firstlast[1][1]
        kwargs = {}
        for var in infile.variables:
            if var in self.ElementType.variables:
                kwargs[var] = self.history[var][
                    np.arange(len(index_of_last)), index_of_last]
        kwargs['ID'] = np.arange(len(elements)) + 1
        self.elements = self.ElementType(**kwargs)
        self.elements_deactivated = self.ElementType()
    else:
        self.history = None
        logger.warning('Not importing history')

    # Remove elements which are scheduled for deactivation
    self.remove_deactivated_elements()

    # Import and apply config settings
    attributes = infile.ncattrs()
    for attr in attributes:
        if attr.startswith('config_'):
            value = infile.getncattr(attr)
            conf_key = attr[7:]
            if value == 'True':
                value = True
            if value == 'False':
                value = False
            if value == 'None':
                value = None
            try:
                self.set_config(conf_key, value)
                logger.debug('Setting imported config: %s -> %s' %
                             (conf_key, value))
            except:
                logger.warning('Could not set config: %s -> %s' %
                                (conf_key, value))

    # Import time steps from metadata
    def timedelta_from_string(timestring):
        if 'day' in timestring:
            days = int(timestring.split('day')[0])
            hs = timestring.split(' ')[-1]
            th = datetime.strptime(hs, '%H:%M:%S')
            return timedelta(days=days, hours=th.hour, minutes=th.minute, seconds=th.second)
        else:
            t = datetime.strptime(timestring, '%H:%M:%S')
            return timedelta(
                hours=t.hour, minutes=t.minute, seconds=t.second)
    try:
        self.time_step = timedelta_from_string(infile.time_step_calculation)
        self.time_step_output = timedelta_from_string(infile.time_step_output)
    except Exception as e:
        logger.warning(e)
        logger.warning('Could not parse time_steps from netCDF file')

    infile.close()



import numpy as np


def vec_nearest(X, xp):
    """
    Find nearest element in vector `X` to `xp`.

    Args:

        X   Array
        xp  Array

    Returns:
        i   M   vector of indices [0..N] of closest element in
                X[0..N, i] to xp[i]
    """
    xp = np.atleast_1d(xp)
    return np.argmin(np.abs(X - xp), axis=0)




"""
Main class that represents an oil record.

This maps to the JSON used in the DB

Having a Python class makes it easier to write importing, validating etc, code.
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json

from ..common.measurement import NeedleAdhesion
from .properties import DispersibilityList, EmulsionList, ESTSEvaporationTest


@dataclass_to_json
@dataclass
class EnvironmentalBehavior:
    dispersibilities: DispersibilityList = field(default_factory=DispersibilityList)
    emulsions: EmulsionList = field(default_factory=EmulsionList)
    adhesion: NeedleAdhesion = None
    ests_evaporation_test: ESTSEvaporationTest = field(default_factory=ESTSEvaporationTest)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Gaute Hope, MET Norway
"""
Interface to the ADIOS oil database.
"""

import logging

logger = logging.getLogger(__name__)

import numpy as np
from typing import List

from .models.oil.oil import Oil as AdiosOil
from .computation import gnome_oil
from .computation import physical_properties
from .util.estimations import oil_water_surface_tension_from_api


class NotFullOil(Exception):
    pass


def __require_gnome_oil__(f):
    def w(self, *args, **kwargs):

        if self.gnome_oil is None:
            raise NotFullOil()

        return f(self, *args, **kwargs)

    return w


class OpendriftOil:
    id: str
    type: str
    name: str
    API: float
    gnome_suitable: bool
    labels: List[str]
    location: str
    model_completeness: float
    product_type: str
    sample_date: str
    data: dict
    oil: AdiosOil
    gnome_oil: dict

    def __init__(self, o):
        self.data = o

        data = o['data']
        meta = data['attributes']['metadata']
        self.id = data['_id']
        self.name = meta['name']

        logger.debug(f'Parsing Oil: {self.id} / {self.name}')
        self.oil = AdiosOil.from_py_json(data['attributes'])

        if not self.oil.metadata.gnome_suitable:
            logger.error(f'{self.id} / {self.name}: is not GNOME suitable')
        else:
            self.gnome_oil = gnome_oil.make_gnome_oil(self.oil)

    def __repr__(self):
        return f"[<adios.Oil> {self.id}] {self.name}"

    def json(self):
        import json
        return json.dumps(self.data)

    def make_full(self) -> 'OpendriftOil':
        return self

    def valid(self):
        """
        Check whether this oil can be used in Opendrift simulations.
        """
        return self.oil.metadata.gnome_suitable

    @__require_gnome_oil__
    def density_at_temp(self, t, unit='K') -> float:
        """
        Return density at temperature (in Kelvin by default).
        """
        return physical_properties.Density(self.oil).at_temp(t, unit)

    @__require_gnome_oil__
    def kvis_at_temp(self, t, unit='K') -> float:
        return physical_properties.KinematicViscosity(self.oil).at_temp(
            t, temp_units=unit)

    @property
    @__require_gnome_oil__
    def mass_fraction(self) -> np.ndarray:
        return np.asarray(self.gnome_oil['mass_fraction'])

    @__require_gnome_oil__
    def oil_water_surface_tension(self) -> float:
        return oil_water_surface_tension_from_api(self.gnome_oil['api'])

    @property
    @__require_gnome_oil__
    def bulltime(self) -> float:
        bulltime = self.gnome_oil['bullwinkle_time']
        if bulltime is None:
            return -999.  # legacy from old oil_library
        else:
            return bulltime

    @property
    @__require_gnome_oil__
    def bullwinkle(self) -> float:
        return self.gnome_oil['bullwinkle_fraction']

    @property
    @__require_gnome_oil__
    def emulsion_water_fraction_max(self) -> float:
        return self.gnome_oil['emulsion_water_fraction_max']

    @__require_gnome_oil__
    def vapor_pressure(self, temp) -> float:
        '''
        Calculate vapor pressure. This method is taken from the old oil_library.

        Args:

            temp: temperature in Kelvin.

        Returns:

            Array of vapor pressures for each component. Pascal.
        '''
        atmos_pressure = 101325.0
        boiling_point = np.asarray(self.gnome_oil['boiling_point'])

        D_Zb = 0.97
        R_cal = 1.987  # calories

        D_S = 8.75 + 1.987 * np.log(boiling_point)
        C_2i = 0.19 * boiling_point - 18

        var = 1. / (boiling_point - C_2i) - 1. / (temp - C_2i)
        ln_Pi_Po = (D_S * (boiling_point - C_2i)**2 /
                    (D_Zb * R_cal * boiling_point) * var)
        Pi = np.exp(ln_Pi_Po) * atmos_pressure

        return Pi

    @property
    @__require_gnome_oil__
    def molecular_weight(self):
        return np.asarray(self.gnome_oil['molecular_weight'])

    @property
    @__require_gnome_oil__
    def k0y(self) -> float:
        return self.gnome_oil['k0y']



"""
Validation of a single oil record

The actual validation is done in the Oil object (and sub-objects)

This just provides some  convenient wrappers around the validate call
"""

from ..oil import Oil

import logging


# Putting these all here so we can keep track more easily
# from .warnings import WARNINGS
from .errors import ERRORS

logger = logging.getLogger(__name__)


def validate_json(oil_json):
    """
    validate a json-compatible-python record

    An Oil object is returned, if it's possible to do so.

    The "status" field is updated in place, with no other alterations of the record
    """

    if "oil_id" not in oil_json:
        raise ValueError(ERRORS["E010"])
    try:
        oil = Oil.from_py_json(oil_json)
    except TypeError as err:
        if "argument: 'oil_id'" in err.args[0]:
            raise TypeError(ERRORS["E011"].format(oil_json["oil_id"]))
        else:
            raise

    oil.reset_validation()

    return oil


def validate(oil):
    """
    validate an Oil object

    oil.status is updated in place -- this is simply a wrapper around
    Oil.reset_validation() -- probably no longer needed
    """
    oil.reset_validation()



"""
Code for estimating various properties that aren't in a oil record.

These can be used to generate complete oil specifications for
modeling, etc.

Note: there is code in here to make various estimations of
    SARA fractionations -- be warned, these are of unceratain
    use, and not very accurate

"""

import numpy as np

# This is  unit conversion -- should be handled by pynucos
# def api_from_density(density):

#     '''
#         Source: Adios2
#     '''
#     kg_m_3 = density

#     return (141.5 / kg_m_3 * 1000.0) - 131.5


def pour_point_from_kvis(ref_kvis, ref_temp_k):
    '''
        Source: Adios2

        If we have an oil kinematic viscosity at a reference temperature,
        then we can estimate what its pour point might be.
    '''
    c_v1 = 5000.0
    ref_kvis = np.array(ref_kvis)
    ref_temp_k = np.array(ref_temp_k)

    T_pp = (c_v1 * ref_temp_k) / (c_v1 - ref_temp_k * np.log(ref_kvis))

    return T_pp


def flash_point_from_bp(temp_k):
    '''
        Source: Reference: Chang A., K. Pashakanti, and Y. Liu (2012),
                           Integrated Process Modeling and Optimization,
                           Wiley Verlag.

    '''
    temp_k = np.array(temp_k)
    return 117.0 + 0.69 * temp_k


def flash_point_from_api(api):
    '''
        Source: Reference: Chang A., K. Pashakanti, and Y. Liu (2012),
                           Integrated Process Modeling and Optimization,
                           Wiley Verlag.

    '''
    api = np.array(api)
    return 457.0 - 3.34 * api


def cut_temps_from_api(api, N=5):
    '''
        Source: Adios2 & Jones R. (1997),
                A Simplified Pseudo-component Oil Evaporation Model,
                Proceedings of the 20th Arctic and Marine Oil Spill Program,
                Vancouver, CA,
                Vol. 1, pp. 43-62

        Generate distillation cut temperatures from the oil's API.
    '''
    T_0 = 457.0 - 3.34 * api
    T_G = 1357.0 - 247.7 * np.log(api)

    return np.array([(T_0 + (T_G * i) / N) for i in range(N)])


def fmasses_flat_dist(f_res=0, f_asph=0, N=5):
    '''
        Generate a flat distribution of N distillation cut fractional masses.
    '''
    return np.array([(1.0 - f_res - f_asph) / N] * N)


def fmasses_from_cuts(f_evap_i):
    '''
        Generate distillation cut fractional masses from the
        cumulative distillation fractions in the cut data.
    '''
    fmass_i = np.array(f_evap_i)
    fmass_i[1:] = np.diff(fmass_i)

    return fmass_i


def resin_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_res = 0.033 * A + 0.00087 * B - 0.74
    f_res = np.clip(f_res, 0.0, 1.0 - f_other)

    return f_res


def asphaltene_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_asph = (0.000014 * A ** 3.0 +
              0.000004 * B ** 2.0 -
              0.18)
    f_asph = np.clip(f_asph, 0.0, 1.0 - f_other)

    return f_asph


def saturates_fraction(density, viscosity, f_other=0.0):
    A = _A_coeff(density)
    B = _B_coeff(density, viscosity)

    f_sat = -2.5 + 76.6 / A + 0.00013 * np.log(B)
    f_sat = np.clip(f_sat, 0.0, 1.0 - f_other)

    return f_sat


def aromatics_fraction(f_res, f_asph, f_sat):
    f_arom = 1.0 - (f_res + f_asph + f_sat)
    f_arom = np.clip(f_arom, 0.0, 1.0) 

    return f_arom


def _A_coeff(density):
    '''
        Source: Fingas empirical formulas that are based upon analysis
                of ESTS oil properties database.

        This is an intermediate calculation for a coefficient to be
        used to generate the mass fractions of an oil.
    '''
    return 10.0 * np.exp(0.001 * density)


def _B_coeff(density, viscosity):
    '''
        Source: Fingas empirical formulas that are based upon analysis
                of ESTC oil properties database.

        This is an intermediate calculation for a coefficient to be
        used to generate the mass fractions of an oil.
    '''
    return 10.0 * np.log(1000.0 * density * viscosity)


def saturate_mol_wt(boiling_point):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.48 and table 2.6

        Note: for this to actually work in every case, we need to limit
              our temperature to:
              - T_i < 1070.0
              - T_i >
              - T_i > 1070.0 - exp(6.98291)  (roughly about == -8.06)
    '''
    T_i = np.clip(np.array(boiling_point),
                  1070.0 - np.exp(6.98291) + 0.00001,
                  1070.0 - 0.00001)
    return (49.677 * (6.98291 - np.log(1070.0 - T_i))) ** (3.0 / 2.0)


def aromatic_mol_wt(boiling_point):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.48 and table 2.6

        Note: for this to actually work in every case, we need to limit
              our temperature to:
              - T_i < 1015.0
              - T_i > 1015.0 - exp(6.911)  (roughly about == 11.76)
    '''
    T_i = np.clip(np.array(boiling_point),
                  1015.0 - np.exp(6.911) + 0.00001,
                  1015.0 - 0.00001)
    return (44.504 * (6.911 - np.log(1015.0 - T_i))) ** (3.0 / 2.0)


def resin_mol_wt(boiling_points):
    '''
        Source: Recommendation from Bill Lehr
    '''
    return np.full_like(boiling_points, 800.0)


def asphaltene_mol_wt(boiling_points):
    '''
        Source: Recommendation from Bill Lehr
    '''
    return np.full_like(boiling_points, 1000.0)


def trial_densities(boiling_points, watson_factor):
    '''
        Source: Dr. M. R. Riazi,
                Characterization and Properties of Petroleum Fractions
                eq. 2.13 and table 9.6

        Generate an initial estimate of volatile oil components based
        on boiling points and the Watson Characterization Factor.
        This is only good for estimating Aromatics & Saturates.
    '''
    rho_i = 1000.0 * (1.8 * boiling_points) ** (1.0 / 3.0) / watson_factor

    return np.clip(np.array(rho_i), 0, 1090)


def saturate_densities(boiling_points):
    K_w_sat = 12.0

    return trial_densities(boiling_points, K_w_sat)


def aromatic_densities(boiling_points):
    K_w_arom = 10.0

    return trial_densities(boiling_points, K_w_arom)


def resin_densities(boiling_points):
    resin_rho =  np.full(len(boiling_points),1100.0)
    return resin_rho


def asphaltene_densities(boiling_points):
    return np.full_like(boiling_points, 1100.0)


def saturate_mass_fraction(fmass_i, temp_k, total_sat=None):
    '''
        Source: Dr. Robert Jones, based on average of 51 Exxon oils
        This assumes we do not known the SARA totals for the oil

    '''
    T_i = temp_k
    k = .0877

    if total_sat is not None:
#         sat_pct_temp = (124.1069 - k * T_i)
#         sat_pct_i = np.clip(sat_pct_temp,0,100)
#         f_sat_i = fmass_i * sat_pct_i / 100.
        k = (124.1069 *fmass_i.sum() - total_sat*100) / (fmass_i*T_i).sum()
        sat_pct_i = 124.1069 - k * T_i
        f_sat_i = fmass_i * sat_pct_i / 100.
#         scale = total_sat / f_sat_i.sum()
#         f_sat_i = scale * f_sat_i
#         sat_pct_i = f_sat_i / fmass_i
    else:
        sat_pct_i = 124.1069 - k * T_i
        f_sat_i = fmass_i * sat_pct_i / 100.

    #sat_pct_i = 124.1069 - k * T_i
    #f_sat_i = fmass_i * sat_pct_i / 100.

    return np.clip(f_sat_i, 0.0, fmass_i)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from datetime import datetime, timedelta
import numpy as np
import pyproj

from opendrift.readers.basereader import BaseReader, ContinuousReader


class Reader(BaseReader, ContinuousReader):
    """Analytical current field with double gyre"""

    def __init__(self, initial_time=datetime(2000,1,1,0,0),
                 epsilon=0.1, omega=0.628, A=0.25,
                 proj4='+proj=stere +lat_0=0 +lon_0=0 +lat_ts=0 ' +
                       '+units=m +a=6.371e+06 +e=0 +no_defs'):

        self.fileName = 'double_gyre'
        self.name = 'double_gyre'

        self.proj4 = proj4
        self.proj = pyproj.Proj(proj4)

        self.xmin = 0.
        self.xmax = 2.
        self.ymin = 0.
        self.ymax = 1.
        self.A = A
        self.epsilon=epsilon
        self.omega=omega
        self.initial_time = initial_time
        self.start_time = None
        self.end_time = None
        self.time_step = None
        self.variables = ['x_sea_water_velocity',
                          'y_sea_water_velocity']

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        requestedVariables, time, x, y, z, outside = self.check_arguments(
            requestedVariables, time, x, y, z)

        variables = {}

        # Construct double gyre current field
        # https://shaddenlab.berkeley.edu/uploads/LCS-tutorial/examples.html
        t = (time-self.initial_time).total_seconds()
        a = self.epsilon*np.sin(self.omega*t)
        b = 1-2*self.epsilon*np.sin(self.omega*t)
        f = a*x*x + b*x
        dfdx = 2*a*x + b

        variables['x_sea_water_velocity'] = -np.pi*self.A*np.sin(np.pi*f)*np.cos(np.pi*y)
        variables['y_sea_water_velocity'] = np.pi*self.A*np.cos(np.pi*f)*np.sin(np.pi*y)*dfdx

        variables['land_binary_mask'] = np.zeros(x.shape)
        variables['time'] = time
        variables['x'] = x
        variables['y'] = y

        return variables



"""
Main class that represents an oil record.

This maps to the JSON used in the DB

Having a Python class makes it easier to write importing, validating etc, code.
"""
from dataclasses import dataclass, field

from .validation.errors import ERRORS

from ..common.utilities import dataclass_to_json, JSON_List

from ..common.measurement import (Temperature,
                                  Density,
                                  DynamicViscosity,
                                  KinematicViscosity,
                                  AngularVelocity,
                                  InterfacialTension)


class RefTempList:
    """
    mixin for all classes that are a list of points with
    reference temperatures
    """

    def validate(self):
        """
        validater for anything that has a list of reference temps

        e.g. density and viscosity

        For viscosity it checks for shear rate as well.
        """
        points_list = self
        data_str = self.__class__.__name__
        msgs = []
        # check for odd temperatures
        for pt in points_list:
            if pt.ref_temp is None:
                msgs.append(ERRORS["E042"]
                            .format(data_str + " reference temp"))
                return msgs

            temp = pt.ref_temp.converted_to('C').value
            if temp is None:
                msgs.append(ERRORS["E042"]
                            .format(data_str + " reference temp"))
                return msgs
            if temp < -100.0:  # arbitrary, but should catch K/C confusion
                t = f"{pt.ref_temp.value:.2f} {pt.ref_temp.unit}"
                msgs.append(ERRORS["E040"].format(data_str, t))

        # check for duplicate temp/shear_rate combos
        temps = []
        for p in points_list:
            temp = p.ref_temp.converted_to('K').value
            try:
                temp = temp + p.shear_rate.value
            except (TypeError, AttributeError):
                pass
            temps.append(temp)
        temps.sort()
        diff = (abs(t2 - t1) for t1, t2 in zip(temps[1:], temps[:1]))
        for d in diff:
            if d < 1e-3:
                msgs.append(ERRORS["E050"].format("Temperatures", data_str))

        return msgs


@dataclass_to_json
@dataclass
class DensityPoint:
    density: Density = None
    ref_temp: Temperature = None
    method: str = None


class DensityList(JSON_List, RefTempList):
    item_type = DensityPoint


@dataclass_to_json
@dataclass
class DynamicViscosityPoint:
    viscosity: DynamicViscosity = None
    ref_temp: Temperature = None
    shear_rate: AngularVelocity = None
    method: str = None


class DynamicViscosityList(JSON_List, RefTempList):
    item_type = DynamicViscosityPoint


@dataclass_to_json
@dataclass
class KinematicViscosityPoint:
    viscosity: KinematicViscosity = None
    ref_temp: Temperature = None
    shear_rate: AngularVelocity = None
    method: str = None


class KinematicViscosityList(JSON_List, RefTempList):
    item_type = KinematicViscosityPoint


@dataclass_to_json
@dataclass
class PourPoint:
    measurement: Temperature = None
    method: str = None


@dataclass_to_json
@dataclass
class FlashPoint:
    measurement: Temperature = None
    method: str = None


@dataclass_to_json
@dataclass
class InterfacialTensionPoint:
    tension: InterfacialTension
    ref_temp: Temperature
    # interface: str = None  # the interface is given in the attribute name
    method: str = None


class InterfacialTensionList(JSON_List, RefTempList):
    item_type = InterfacialTensionPoint


@dataclass_to_json
@dataclass
class PhysicalProperties:
    pour_point: PourPoint = None
    flash_point: FlashPoint = None

    densities: DensityList = field(default_factory=DensityList)
    kinematic_viscosities: KinematicViscosityList = field(default_factory=KinematicViscosityList)
    dynamic_viscosities: DynamicViscosityList = field(default_factory=DynamicViscosityList)

    interfacial_tension_air: InterfacialTensionList = field(default_factory=InterfacialTensionList)
    interfacial_tension_water: InterfacialTensionList = field(default_factory=InterfacialTensionList)
    interfacial_tension_seawater: InterfacialTensionList = field(default_factory=InterfacialTensionList)



from abc import abstractmethod
import numpy as np
import logging
logger = logging.getLogger(__name__)


class Reader:
    @abstractmethod
    def variables(self):
        """
        Get list of variables
        """
        return []

    @abstractmethod
    def read_grid(self, grid, var, t):
        """
        Read variable for grid.

        Args:

            grid: grid to read variable for

            var: list of strs, name of variables to read

            t: datetime (UTC)

        Returns:

            v0, v1, ...

            tuple with each variables specified in var with same shape as `grid`.
        """
        pass


class ConstantReader(Reader):
    def __init__(self, consts):
        """
        Args:

            consts: dict, name of var: constant
        """
        logger.info("constant reader for: %s" % ", ".join(consts.keys()))
        self.consts = consts
        super().__init__()

    def read_grid(self, grid, var, _):
        for v in var:
            assert v in self.consts, "missing variable %s" % v

        return tuple(self.consts[v] * np.ones(grid.grid.shape) for v in var)

    def variables(self):
        return list(self.consts.keys())

    @staticmethod
    def new_xy(x=.5, y=.5):
        return ConstantReader({
            'x_sea_water_velocity': x,
            'y_sea_water_velocity': y
        })


class OpendriftReader(Reader):
    """
    Wrapper around an Opendrift reader.
    """
    def __init__(self, reader):
        """
        Args:

            reader: opendrift reader
        """
        logger.info("opendrift reader: %s (%s)" %
                    (reader.name, ", ".join(reader.variables)))

        self.r = reader

        super().__init__()

    def variables(self):
        return self.r.variables

    def read_grid(self, grid, var, t):
        logger.debug('reading %s for %s' % (var, t))

        x, y = self.r.lonlat2xy(grid.lons.ravel(), grid.lats.ravel())

        env, _ = self.r.get_variables_interpolated_xy(
            variables=var,
            time=t,
            x=x,
            y=y,
            z=np.zeros(grid.grid.shape).ravel(),
            rotate_to_proj=grid.srs)

        u = tuple(env[v].reshape(grid.grid.shape).filled(fill_value=np.nan) for v in var)

        for uu, vv in zip(u, var):
            if np.isnan(uu).any():
                logger.warning("nan's in %s" % vv)

        return u



"""
Code for making a "GnomeOil" from an Oil Object

See the PyGNOME code for more about GNOME's requirements

NOTE: This make s JSON compatible Python structure from which to build
a GnomeOil
"""

import copy

import numpy as np

import unit_conversion as uc

import logging
logger = logging.getLogger(__name__)

from . import estimations as est
from .physical_properties import get_density_data, get_kinematic_viscosity_data, get_distillation_cuts
from .physical_properties import bullwinkle_fraction, max_water_fraction_emulsion
from .physical_properties import Density, KinematicViscosity
from .estimations import pour_point_from_kvis, flash_point_from_bp, flash_point_from_api


def get_empty_dict():
    """
    This provides an empty dictionary with everything that is needed
    to generate a GNOME Oil
    """
    return{"name": "",
           # Physical properties
           "api": None,
           "pour_point": None,
           "solubility": None,  # kg/m^3
           # emulsification properties
           "bullwinkle_fraction": None,
           "bullwinkle_time": None,
           "emulsion_water_fraction_max": None,
           "densities": [],
           "density_ref_temps": [],
           "density_weathering": [],
           "kvis": [],  # m/s^2
           "kvis_ref_temps": [],  # K
           "kvis_weathering": [],
           # PCs:
           "mass_fraction": [],
           "boiling_point": [],
           "molecular_weight": [],
           "component_density": [],
           "sara_type": [],
           "flash_point": None,
           "adios_oil_id": None,
           }


def make_gnome_oil(oil):
    """
    Make a dict that a GnomeOil can be built from

    A GnomeOil needs:

              "name,
              "# Physical properties
              "api,
              "pour_point,
              "solubility,  # kg/m^3
              "# emulsification properties
              "bullwinkle_fraction,
              "bullwinkle_time,
              "emulsion_water_fraction_max,
              "densities,
              "density_ref_temps,
              "density_weathering,
              "kvis,
              "kvis_ref_temps,
              "kvis_weathering,
              "# PCs:
              "mass_fraction,
              "boiling_point,
              "molecular_weight,
              "component_density,
              "sara_type,
              "flash_point=None,
              "adios_oil_id=None,

    """
    # metadata:
    go = get_empty_dict()
    # make sure we don't change the original oil object
    # oil = copy.deepcopy(oil)
    go['name'] = oil.metadata.name

    dens = Density(oil)
    ref_density = dens.at_temp(288.7)  # 60F in K
    go['api'] = uc.convert('kg/m^3', 'API', ref_density)
    # for gnome_oil we don't treat api as data, only api from density
    oil.metadata.API = go['api']
    go['adios_oil_id'] = oil.oil_id

    # Physical properties
    phys_props = oil.sub_samples[0].physical_properties

    flash_point = phys_props.flash_point
    if flash_point is None:
        go['flash_point'] = None
    else:
        fp = phys_props.flash_point.measurement.converted_to('K')
        if fp.max_value is not None:
            go['flash_point'] = fp.max_value
        elif fp.value is not None:
            go['flash_point'] = fp.value
        else:  # shouldn't happen, but ...
            go['flash_point'] = None

    pour_point = phys_props.pour_point
    if pour_point is None:
        go['pour_point'] = estimate_pour_point(oil)
    else:
        pp = phys_props.pour_point.measurement.converted_to('K')
        if pp.max_value is not None:
            go['pour_point'] = pp.max_value
        elif pp.value is not None:
            go['pour_point'] = pp.value
        elif pp.min_value is not None:
            go['pour_point'] = pp.min_value
        else:
            go['pour_point'] = None

    # fixme: We need to get the weathered densities, if they are there.
    densities = get_density_data(oil, units="kg/m^3", temp_units="K")

    go['densities'], go['density_ref_temps'] = zip(*densities)
    go['density_weathering'] = [0.0] * len(go['densities'])

    viscosities = get_kinematic_viscosity_data(oil, units="m^2/s", temp_units="K")

    if viscosities:
        go['kvis'], go['kvis_ref_temps'] = zip(*viscosities)
        go['kvis_weathering'] = [0.0] * len(go['kvis'])
    else:
        raise ValueError("Gnome oil needs at least one viscosity value")

    bullwinkle = None
    for sub_sample in oil.sub_samples:
        try:
            frac_weathered = sub_sample.metadata.fraction_weathered.converted_to('fraction').value
            if bullwinkle is None or frac_weathered > bullwinkle:
                bullwinkle = frac_weathered
        except:
            frac_weathered = None

    if bullwinkle is None:
        go['bullwinkle_fraction'] = bullwinkle_fraction(oil)
    else:
        go['bullwinkle_fraction'] = bullwinkle

    #go['bullwinkle_fraction'] = bullwinkle_fraction(oil)
    go['emulsion_water_fraction_max'] = max_water_fraction_emulsion(oil)
    go['solubility'] = 0
    go['k0y'] = 2.024e-06 #do we want this included?

    # pseudocomponents
    cut_temps, frac_evap = normalized_cut_values(oil)

    mass_fraction = component_mass_fractions(oil)
    mask = np.where(mass_fraction==0)
    mol_wt = np.delete(component_mol_wt(cut_temps),mask)
    comp_dens = np.delete(component_densities(cut_temps),mask)
    boiling_pt = np.delete(component_temps(cut_temps),mask)
    sara_type = np.delete(np.array(component_types(cut_temps)),mask)
    mass_frac = np.delete(mass_fraction,mask)

    go['molecular_weight'] = mol_wt.tolist()
    go['component_density'] = comp_dens.tolist()
    go['mass_fraction'] = mass_frac.tolist()
    go['boiling_point'] = boiling_pt.tolist()
    go['sara_type'] = sara_type.tolist()

    return go


def estimate_pour_point(oil):
    """
    estimate pour point from kinematic viscosity

    """

    pour_point = None
    kvis = get_kinematic_viscosity_data(oil)
    c_v1 = 5000.0
    if kvis:
        lowest_kvis = kvis[0]

        #pour_point = pour_point_from_kvis(lowest_kvis[0], lowest_kvis[1])
        pour_point = (c_v1 * lowest_kvis[1]) / (c_v1 - lowest_kvis[1] * np.log(lowest_kvis[0]))

    return pour_point


def estimate_flash_point(oil):
    """
    estimate flash point from api or boiling point

    """

    oil_api = oil.metadata.API

    cuts = get_distillation_cuts(oil)

    # fixme: if we do need this, we should have a better way to get
    #        boiling point -- the first cut is not necessarily the BP!
    #        the IBP is stored in the distillation record, if it is known.
    if len(cuts) > 2:
        lowest_cut = cuts[0]
        flash_point = flash_point_from_bp(lowest_cut[1])
    elif oil_api is not None:
        flash_point = flash_point_from_api(oil_api)
    else:
        flash_point = None

    return flash_point


def component_temps(cut_temps, N=10):
    """
    component temps from boiling point

    """
    component_temps = np.append(cut_temps, list(zip(cut_temps, cut_temps, cut_temps, cut_temps)))
    len_ct = len(cut_temps)
    new_temps = component_temps[len_ct:].copy()

    return np.asarray(new_temps)

def component_types(cut_temps, N=10):
    """
    set component SARA types

    """
    T_i = component_temps(cut_temps, N)

    types_out = ['Saturates', 'Aromatics', 'Resins', 'Asphaltenes'] * int((len(T_i) / 4))

    return types_out


def component_densities(boiling_points):
    """
    estimate component densities from boiling point

    """
    rho_list = list(zip(est.saturate_densities(boiling_points),
                  est.aromatic_densities(boiling_points),
                  est.resin_densities(boiling_points),
                  est.asphaltene_densities(boiling_points)))

    return (np.asarray(rho_list)).flatten()


def component_mol_wt(boiling_points):
    """
    estimate component molecular weight from boiling point

    """
    mw_list = list(zip(est.saturate_mol_wt(boiling_points),
                  est.aromatic_mol_wt(boiling_points),
                  est.resin_mol_wt(boiling_points),
                  est.asphaltene_mol_wt(boiling_points)))

    return (np.asarray(mw_list)).flatten()


def inert_fractions(oil, density=None, viscosity=None):
    """
    resins and asphaltenes from database or estimated if None

    """
    f_res = None
    f_asph = None
    estimated_res = estimated_asph = False

    f_res = oil.sub_samples[0].SARA.resins
    f_asph = oil.sub_samples[0].SARA.asphaltenes

    if f_res is not None:
        f_res = f_res.converted_to('fraction').value
    if f_asph is not None:
        f_asph = f_asph.converted_to('fraction').value

    if f_res is not None and f_asph is not None:
        return f_res, f_asph, estimated_res, estimated_asph
    else:
        dens = Density(oil)
        density = dens.at_temp(288.15)
        kvis = KinematicViscosity(oil)
        viscosity = kvis.at_temp(288.15)

    if f_res is None:
        f_res = est.resin_fraction(density, viscosity)
        estimated_res = True

    if f_asph is None:
        f_asph = est.asphaltene_fraction(density, viscosity, f_res)
        estimated_asph = True

    return f_res, f_asph, estimated_res, estimated_asph

#this will all be removed...
def _linear_curve(x, a, b):
    '''
        Here we describe the form of a linear function for the purpose of
        curve-fitting measured data points.
    '''
    return (a * x + b)


def clamp(x, M, zeta=0.03):
    '''
        We make use of a generalized logistic function or Richard's curve
        to generate a linear function that is clamped at x == M.
        We make use of a zeta value to tune the parameters nu, resulting in a
        smooth transition as we cross the M boundary.
    '''
    return (x -
            (x / (1.0 + np.e ** (-15 * (x - M))) ** (1.0 / (1 + zeta))) +
            (M / (1.0 + np.e ** (-15 * (x - M))) ** (1.0 / (1 - zeta))))


def _inverse_linear_curve(y, a, b, M, zeta=0.12):
    y_c = clamp(y, M, zeta)

    return (y_c - b) / a


# adios version
def normalized_cut_values_adios(oil):
    oil_api = oil.metadata.API
    T0=457.16-3.3447*oil_api
    TG=1356.7-247.36*np.log(oil_api)
    cuts = get_distillation_cuts(oil)
    if len(cuts) == 0:
        # should be a warning if api < 50 or not a crude
        #NumComp=5					 #number of components
        NumComp=10					 #number of components
        BP = np.array([(T0 + (TG *(i+.5)) / NumComp) for i in range(NumComp)])
        fevap = np.cumsum(est.fmasses_flat_dist(N=10))
    else:
        NumCuts = len(cuts)
        NumComp = NumCuts + 1
        BP_i, fevap_i = list(zip(*[(c[1], c[0]) for c in cuts]))
        BP = np.zeros(NumComp)
        fevap = np.zeros(NumComp)
        len_BP = len(BP)
        T1 = BP_i[0]
        T2 = BP_i[1]
        TN = BP_i[NumCuts-1]
        fN = fevap_i[NumCuts-1]
        BP[0] = T1-.5*fevap_i[0]*(T2-T1)/(fevap_i[1]-fevap_i[0])
        fevap[0] = fevap_i[0]
        for i in range(NumCuts-1):
            BP[i+1] = .5*(BP_i[i]+BP_i[i+1])
            fevap[i+1] = fevap_i[i+1]
        BPN =  TN+.5*(1-fN)*(TN-T1)/(fN-fevap_i[0])
        fevap[NumComp-1] = 1
        BP[NumComp-1] = TN+.5*(1-fN)*(TN-T1)/(fN-fevap_i[0])
        if BP[NumComp-1]> 1500.:
            BP[NumComp-1] = 1500
        if BP[NumComp-1]< 1050:
            BP[NumComp-1] = 1050

    return np.asarray(BP), est.fmasses_from_cuts(fevap)

def normalized_cut_values(oil):
    f_res = f_asph = 0 # for now, we are including the resins and asphaltenes
    cuts = get_distillation_cuts(oil)
    oil_api = oil.metadata.API
    iBP = 266
    tBP = 1050

    if len(cuts) == 0:
        # should be a warning if api < 50 or not a crude
        oil_api = oil.metadata.API
        if oil.metadata.product_type != 'Crude Oil NOS':
            logger.error('warning: oil not recommended for use in Gnome')
            # print(WARNINGS['W007'] + "  - oil not recommended for use in Gnome")
        if oil_api < 0:
            raise ValueError("Density is too large for estimations. Oil not suitable for use in Gnome")

        # ADIOS2 method
        BP_i = est.cut_temps_from_api(oil_api)
        fevap_i = np.cumsum(est.fmasses_flat_dist(f_res, f_asph))
        # Robert's new method
        #iBP = 10/9 * (519.3728 - 3.6637 * oil_api) - 1015 / 9
        #tBP = 1015
        #BP_i = [iBP, tBP]
        #fevap_i = [0,1]
    else:
        BP_i, fevap_i = list(zip(*[(c[1], c[0]) for c in cuts]))
        N = len(BP_i)
        if not (fevap_i[1] == fevap_i[0]):
            iBP = BP_i[0] - fevap_i[0] * (BP_i[1] - BP_i[0]) / (fevap_i[1] - fevap_i[0])
        if not (fevap_i[N-1] == fevap_i[0]):
            tBP = BP_i[N-1] + (1 - fevap_i[0]) * (BP_i[N-1] - BP_i[0]) / (fevap_i[N-1] - fevap_i[0])

    iBP = max(266, iBP)
    tBP = min(1050, tBP)

    set_temp = [266,310,353,483,563,650,800,950,1050]
    #set_temp = [266,310,353,423,483,523,543,563,650,800,950,1050]

    N = len(BP_i)

    new_fevap = fevap_i
    new_BP = BP_i
    if not fevap_i[N-1] == 1:
        new_BP = np.append(BP_i,tBP)
        new_fevap = np.append(fevap_i, 1.0)
    if not fevap_i[0] == 0:
        new_BP = np.insert(new_BP,0,iBP)
        new_fevap = np.insert(new_fevap, 0, 0)
    new_evap = np.interp(set_temp, new_BP, new_fevap)

    if new_evap[-1] < 1:
        new_evap[-1] = 1	# put all the extra mass in the last cut

    avg_evap_i = np.asarray(new_evap[1:])
    avg_temp_i = np.asarray([(set_temp[i]+set_temp[i+1])/2 for i in range(0, len(set_temp)-1)])

    avg_temp_i = avg_temp_i[avg_evap_i!=0]
    avg_evap_i = avg_evap_i[avg_evap_i!=0]

    num_ones = 0
    for i in range(len(avg_evap_i)):
        if avg_evap_i[i] == 1:
            num_ones += 1

    if num_ones > 1:
        for i in range(num_ones-1):
            avg_temp_i = np.delete(avg_temp_i,len(avg_evap_i)-1)
            avg_evap_i = np.delete(avg_evap_i,len(avg_evap_i)-1)

    return np.asarray(avg_temp_i), est.fmasses_from_cuts(avg_evap_i)

def normalized_cut_values_james(oil, N=10):
    """
    estimate cut temperatures

    """
    from scipy.optimize import curve_fit	#temporary
    #f_res, f_asph, _estimated_res, _estimated_asph = inert_fractions(oil)
    f_res = f_asph = 0
    cuts = get_distillation_cuts(oil)
    oil_api = oil.metadata.API
    if len(cuts) == 0:
        if oil.metadata.product_type != 'Crude Oil NOS':
            logger.error('warning: oil not recommended for use in Gnome')
            # print(WARNINGS['W007'] + "  - oil not recommended for use in Gnome")
        if oil_api < 0:
            raise ValueError("Density is too large for estimations. Oil not suitable for use in Gnome")
        BP_i = est.cut_temps_from_api(oil_api)
        fevap_i = np.cumsum(est.fmasses_flat_dist(f_res, f_asph))
    else:
        BP_i, fevap_i = list(zip(*[(c[1], c[0]) for c in cuts]))

    popt, _pcov = curve_fit(_linear_curve, BP_i, fevap_i)
    f_cutoff = _linear_curve(732.0, *popt)  # center of asymptote (< 739)
    popt = popt.tolist() + [f_cutoff]

    fevap_i = np.linspace(0.0, 1.0 - f_res - f_asph, (N * 2) + 1)[1:]
    T_i = _inverse_linear_curve(fevap_i, *popt)

    fevap_i = fevap_i.reshape(-1, 2)[:, 1]
    T_i = T_i.reshape(-1, 2)[:, 0]

    above_zero = T_i > 0.0
    T_i = T_i[above_zero]
    fevap_i = fevap_i[above_zero]

    return T_i, est.fmasses_from_cuts(fevap_i)

# need to replace this
def component_mass_fractions(oil):
    """
    estimate pseudocomponent mass fractions

    """
    cut_temps, fmass_i = normalized_cut_values(oil)
    measured_sat = oil.sub_samples[0].SARA.saturates
    sat, arom, res, asph = sara_totals(oil)
    if measured_sat is not None:
        f_sat_i = est.saturate_mass_fraction(fmass_i,cut_temps,sat)	#scale by data value
    else:
        f_sat_i = est.saturate_mass_fraction(fmass_i,cut_temps)
    f_arom_i = fmass_i - f_sat_i

    f_res_i = np.zeros_like(f_sat_i)
    f_asph_i = np.zeros_like(f_sat_i)

    if asph + res > f_arom_i.sum(): # something went wrong
        res = 0
        asph = 0

    f_res_i[len(f_res_i)-1] = res
    f_asph_i[len(f_asph_i)-1] = asph
    f_inert = asph + res

    for i in range(len(f_arom_i)-1,-1,-1):
        if f_inert > 0:
            if f_arom_i[i] > f_inert:
                f_arom_i[i] = f_arom_i[i] - f_inert
                f_inert = 0
            else:
                f_inert = f_inert - f_arom_i[i]
                f_arom_i[i] = 0
        else:
            f_inert = f_inert * 1

    mf_list = list(zip(f_sat_i, f_arom_i, f_res_i, f_asph_i))	# will want to zip all 4 together
    return (np.asarray(mf_list)).flatten()

def sara_totals(oil):
    """
    get SARA from database
    estimate if no data

    """
    SARA = oil.sub_samples[0].SARA
    aromatics = oil.sub_samples[0].SARA.aromatics
    saturates = oil.sub_samples[0].SARA.saturates
    resins = oil.sub_samples[0].SARA.resins
    asphaltenes = oil.sub_samples[0].SARA.asphaltenes

    dens = Density(oil)
    density = dens.at_temp(288.15)
    kvis = KinematicViscosity(oil)
    viscosity = kvis.at_temp(288.15)

    if resins is None:
        resins_total = est.resin_fraction(density, viscosity)
    else:
        resins_total = resins.converted_to('fraction').value
    if asphaltenes is None:
        asphaltenes_total = est.asphaltene_fraction(density, viscosity, resins_total)
    else:
        asphaltenes_total = asphaltenes.converted_to('fraction').value
    if saturates is None:
        saturates_total = est.saturates_fraction(density, viscosity)
    else:
        saturates_total = saturates.converted_to('fraction').value
    if aromatics is None:
        aromatics_total = est.aromatics_fraction(resins_total,asphaltenes_total,saturates_total)
    else:
        aromatics_total = aromatics.converted_to('fraction').value

    #print("SARA totals = ", saturates_total,aromatics_total,resins_total, asphaltenes_total)

    return saturates_total, aromatics_total, resins_total, asphaltenes_total



#! /usr/bin/env python
#
# Author: Gaute Hope (gauteh@met.no) / 2022
#
# Converts oils from old CSV format to JSON format using `adios_db` (make sure it is installed from adios_oil_database).
#
# Usage:
#
#   csv_to_json.py CSVFILE outdir/

import sys
import logging
import coloredlogs
from pathlib import Path

from adios_db.data_sources.noaa_fm import OilLibraryCsvFile, OilLibraryRecordParser, OilLibraryAttributeMapper
from adios_db.models.oil.validation.validate import validate_json
from adios_db.models.oil.completeness import set_completeness

logger = logging.getLogger(__name__)
coloredlogs.install('debug')

CSV = sys.argv[1]
OUT = sys.argv[2]

logger.info(f"Reading input file: {CSV}")
logger.info(f"Saving JSON files to: {OUT}")

config = None
reader = OilLibraryCsvFile
parser = OilLibraryRecordParser
mapper = OilLibraryAttributeMapper

for record in reader(CSV).get_records():
    oil_map = mapper(parser(*record))
    oil_json = oil_map.py_json()
    oil = validate_json(oil_json)
    set_completeness(oil)

    outf = Path(OUT).joinpath(f"{oil.oil_id}.json")
    logger.info(f"Writing {outf}..")
    with open(outf, 'w') as fd:
        oil.to_file(outf)



"""
Opendrift module

.. currentmodule:: opendrift

"""
import logging; logger = logging.getLogger(__name__)
import unittest
import importlib
import platform
import numpy as np
import time
from datetime import timedelta
from .version import __version__

# For automated access to available drift classes, e.g. for GUI
# Hardcoded for now
_available_models = \
    ['leeway.Leeway',
     'openoil.OpenOil',
     'larvalfish.LarvalFish',
     'plastdrift.PlastDrift',
     'shipdrift.ShipDrift',
     'openberg.OpenBerg']

def get_model_names():
    return [m.split('.')[-1] for m in _available_models]

def get_model(model_name):
    if model_name not in get_model_names():
        raise ValueError('No drift model named %s' % model_name)
    else:
        for m in _available_models:
            if m.split('.')[-1] == model_name:
                module = importlib.import_module(
                            'opendrift.models.' + m.split('.')[0])
                model = getattr(module, model_name)
                return model


def open(filename, times=None, elements=None, load_history=True):
    '''Import netCDF output file as OpenDrift object of correct class'''

    import os
    import pydoc
    from netCDF4 import Dataset
    if not os.path.exists(filename):
        logger.info('File does not exist, trying to retrieve from URL')
        import urllib
        try:
            urllib.urlretrieve(filename, 'opendrift_tmp.nc')
            filename = 'opendrift_tmp.nc'
        except:
            raise ValueError('%s does not exist' % filename)
    n = Dataset(filename)
    try:
        module_name = n.opendrift_module
        class_name = n.opendrift_class
    except:
        raise ValueError(filename + ' does not contain '
                         'necessary global attributes '
                         'opendrift_module and opendrift_class')
    n.close()

    if class_name == 'OpenOil3D':
        class_name = 'OpenOil'
        module_name = 'opendrift.models.openoil'
    if class_name == 'OceanDrift3D':
        class_name = 'OceanDrift'
        module_name = 'opendrift.models.oceandrift'
    cls = pydoc.locate(module_name + '.' + class_name)
    if cls is None:
        from opendrift.models import oceandrift
        cls = oceandrift.OceanDrift
    o = cls()
    o.io_import_file(filename, times=times, elements=elements, load_history=load_history)
    logger.info('Returning ' + str(type(o)) + ' object')
    return o

def open_xarray(filename, chunks={'trajectory': 50000, 'time': 1000}):
    '''Import netCDF output file as OpenDrift object of correct class'''

    import os
    import pydoc
    import xarray as xr
    if not os.path.exists(filename):
        logger.info('File does not exist, trying to retrieve from URL')
        import urllib
        try:
            urllib.urlretrieve(filename, 'opendrift_tmp.nc')
            filename = 'opendrift_tmp.nc'
        except:
            raise ValueError('%s does not exist' % filename)
    n = xr.open_dataset(filename)
    try:
        module_name = n.opendrift_module
        class_name = n.opendrift_class
    except:
        raise ValueError(filename + ' does not contain '
                         'necessary global attributes '
                         'opendrift_module and opendrift_class')
    n.close()

    if class_name == 'OpenOil3D':
        class_name = 'OpenOil'
        module_name = 'opendrift.models.openoil'
    if class_name == 'OceanDrift3D':
        class_name = 'OceanDrift'
        module_name = 'opendrift.models.oceandrift'
    cls = pydoc.locate(module_name + '.' + class_name)
    if cls is None:
        from opendrift.models import oceandrift
        cls = oceandrift.OceanDrift
    o = cls()
    o.io_import_file_xarray(filename, chunks=chunks)

    logger.info('Returning ' + str(type(o)) + ' object')
    return o

def versions():
    import multiprocessing
    import platform
    import scipy
    import matplotlib
    import netCDF4
    import xarray
    import sys
    s = '\n------------------------------------------------------\n'
    s += 'Software and hardware:\n'
    s += '  OpenDrift version %s\n' % __version__
    s += '  Platform: %s, %s\n' % (platform.system(), platform.release())
    try:
        from psutil import virtual_memory
        ram = virtual_memory().total/(1024**3)
    except:
        ram = 'unknown'
    s += '  %s GB memory\n' % ram
    s += '  %s processors (%s)\n' % (multiprocessing.cpu_count(),
                                   platform.processor())
    s += '  NumPy version %s\n' % np.__version__
    s += '  SciPy version %s\n' % scipy.__version__
    s += '  Matplotlib version %s\n' % matplotlib.__version__
    s += '  NetCDF4 version %s\n' % netCDF4.__version__
    s += '  Xarray version %s\n' % xarray.__version__
    s += '  Python version %s\n' % sys.version.replace('\n', '')
    s += '------------------------------------------------------\n'
    return s


def import_from_ladim(ladimfile, romsfile):
    """Import Ladim output file as OpenDrift simulation obejct"""

    from models.oceandrift import OceanDrift
    o = OceanDrift()
    from netCDF4 import Dataset, date2num, num2date
    if isinstance(romsfile, str):
        from opendrift.readers import reader_ROMS_native
        romsfile = reader_ROMS_native.Reader(romsfile)
    l = Dataset(ladimfile, 'r')
    pid = l.variables['pid'][:]
    particle_count = l.variables['particle_count'][:]
    end_index = np.cumsum(particle_count)
    start_index = np.concatenate(([0], end_index[:-1]))
    x = l.variables['X'][:]
    y = l.variables['Y'][:]
    lon, lat = romsfile.xy2lonlat(x, y)
    time = num2date(l.variables['time'][:],
                    l.variables['time'].units)

    history_dtype_fields = [
        (name, o.ElementType.variables[name]['dtype'])
        for name in o.ElementType.variables]
    # Add environment variables
    o.history_metadata = o.ElementType.variables.copy()
    history_dtype = np.dtype(history_dtype_fields)

    num_timesteps = len(time)
    num_elements = len(l.dimensions['particle'])
    o.history = np.ma.array(
        np.zeros([num_elements, num_timesteps]),
        dtype=history_dtype, mask=[True])

    for n in range(num_timesteps):
        start = start_index[n]
        active = pid[start:start+particle_count[n]]
        o.history['lon'][active, n] = \
            lon[start:start+particle_count[n]]
        o.history['lat'][active, n] = \
            lat[start:start+particle_count[n]]
        o.history['status'][active, n] = 0

    o.status_categories = ['active', 'missing_data']
    firstlast = np.ma.notmasked_edges(o.history['status'], axis=1)
    index_of_last = firstlast[1][1]
    o.history['status'][np.arange(len(index_of_last)),
                        index_of_last] = 1
    kwargs = {}
    for var in ['lon', 'lat', 'status']:
        kwargs[var] = o.history[var][
            np.arange(len(index_of_last)), index_of_last]
    kwargs['ID'] = range(num_elements)
    o.elements = o.ElementType(**kwargs)
    o.elements_deactivated = o.ElementType()
    o.remove_deactivated_elements()
    # Import time steps from metadata
    o.time_step = time[1] - time[0]
    o.time_step_output = o.time_step
    o.start_time = time[0]
    o.time = time[-1]
    o.steps_output = num_timesteps

    return o




__version__ = "1.9.1rc2"


def git_describe():
    """
    Return git version if available.
    """
    import os.path
    from subprocess import check_output, DEVNULL

    path = os.path.dirname(__file__)
    args = [
        "git", "-C", path, "describe", "--tags", "--abbrev=7", "--dirty",
        "--broken"
    ]

    try:
        version = check_output(args, cwd=path, stderr=DEVNULL).decode().strip()
        return version
    except:
        return None

def version_or_git():
    v = git_describe()
    if v is None:
        return __version__
    else:
        return "%s / %s" % (__version__, v)




import numpy as np
from scipy import ndimage

# TODO: Check out `findiff` package.

def grad2d(f, dx, dy):
  return (
      ndimage.gaussian_filter1d(f, sigma = 1, axis = 0, order = 1, mode = 'constant', cval = 0.) / dx,
      ndimage.gaussian_filter1d(f, sigma = 1, axis = 1, order = 1, mode = 'constant', cval = 0.) / dy)




#
# Model class definitions for embedded content in our oil records
#
# Note: These are deprecated a bit.  An updated SARAFraction is sitting in
#       models.oil

from enum import Enum


class ProductTypeEnum(Enum):
    crude = 'crude'
    refined = 'refined'


class SaraTypeEnum(Enum):
    saturates = 'Saturates'
    aromatics = 'Aromatics'
    resins = 'Resins'
    asphaltenes = 'Asphaltenes'


class ToxicityTypeEnum(Enum):
    ec = 'EC'  # effective concentration
    lc = 'LC'  # lethal concentration


# for use in the Emulsion related models
class VisualStabilityEnum(Enum):
    entrained = 'Entrained'
    did_not_form = 'Did not form'
    unstable = 'Unstable'
    stable = 'Stable'
    meso_stable = 'Meso-stable'


# for use in the interfacial tension related models
class InterfaceTypeEnum(Enum):
    air = 'air'
    water = 'water'
    seawater = 'seawater'



#!/usr/bin/env python
#
# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway


# Utility script to display contents of a netCDF CF-compliant
# file or URL containing driver data suitable for opendrift
#
# Knut-Frode Dagestad, 19 Feb 2015

import sys
import argparse
from datetime import datetime
import numpy as np

try:
    from opendrift.readers import reader_netCDF_CF_generic
    from opendrift.readers import reader_netCDF_CF_unstructured
    from opendrift.readers import reader_ROMS_native

    readers = [reader_netCDF_CF_generic, reader_ROMS_native, reader_netCDF_CF_unstructured]
    try:
        from opendrift.readers import reader_grib
        readers.append(reader_grib)
    except:
        pass

except ImportError: # development
    sys.exit('Please add opendrift folder to your PYTHONPATH.')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename',
                        help='<URL or netCDF filename>')
    parser.add_argument('-p', dest='variable',
                        default='noplot', nargs='?',
                        help='Plot domain (or variable if given)')
    parser.add_argument('-vmin', dest='vmin',
                        default=None, nargs='?',
                        help='Minimum value for colorbar')
    parser.add_argument('-vmax', dest='vmax',
                        default=None, nargs='?',
                        help='Maximum value for colorbar')
    parser.add_argument('-e', action='store_true',
                        help='Report errors on failure.')
    parser.add_argument('-lon', dest='lon', default=None,
                        help='Report data from position.')
    parser.add_argument('-lat', dest='lat', default=None,
                        help='Report data from position.')
    parser.add_argument('-time', dest='time', default=None,
                        help='Report data from position at time [YYYYmmddHHMM].')

    args = parser.parse_args()

    for reader in readers:
        try:
            print('Testing %s...' % reader.__file__)
            r = reader.Reader(args.filename)
            print(r)
            break
        except Exception as me:
            if args.e is True:
                print(me)
                import traceback
                print(traceback.format_exc())
                print('---------------------------------------')
            print('...not applicable.')

    if not 'r' in locals():
        sys.exit('No readers applicable for ' + args.filename)

    if args.lon is not None:
        if args.lon is None or args.time is None:
            raise ValueError('Both lon, lat and time must be given')
        lon = np.atleast_1d(float(args.lon))
        lat = np.atleast_1d(float(args.lat))
        x,y = r.lonlat2xy(lon, lat)
        time = datetime.strptime(args.time, '%Y%m%d%H%M')
        r.buffer=3
        i=3; j=3  # center of block
        variables = [var for var in r.variables if
                     var not in ('time') and 'time' not in var]
        data = r.get_variables(variables, time, x, y, z=0)
        for var in variables:
            if data[var].ndim == 2:
                print('%s : %s' % (var, data[var][i,j]))
            elif data[var].ndim == 3:
                print('%s : %s' % (var, data[var][0, i,j]))
            else:
                pass
        if 'x_wind' in variables and 'y_wind' in variables:
            print('windspeed : %s' % np.sqrt(
                data['x_wind'][i,j]**2 + data['y_wind'][i,j]**2))

    if args.variable != 'noplot':
        if args.variable is None:
            r.plot()
        else:
            if args.vmin is None:
                vmin = None
            else:
                vmin = float(args.vmin)
            if args.vmax is None:
                vmax = None
            else:
                vmax = float(args.vmax)

            r.plot(args.variable, vmin=vmin, vmax=vmax)

if __name__ == '__main__':
    main()




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from datetime import datetime, timedelta

import numpy as np
import pyproj

from opendrift.readers.basereader import BaseReader, ContinuousReader


class Reader(BaseReader, ContinuousReader):
    variables = ['x_sea_water_velocity', 'y_sea_water_velocity']

    def __init__(self, lons, lats, times, wind=None, waves=None, z=0,
                 name='reader_current_from_drifter'):
        """Reconstruct currents from time series of drifter positions"""

        self.name = name
        lons = np.asarray(lons)
        lats = np.asarray(lats)
        valid = np.isfinite(lons+lats)
        if np.sum(valid) < len(lons):
            print('Skipping %s invalid positions' % (len(lons)-np.sum(valid)))
            lons = lons[valid]
            lats = lats[valid]
            times = times[valid]

        # Cover whole earth, no validity radius yet
        self.proj4 = '+proj=latlong'
        self.xmin = -180
        self.xmax = 180
        self.ymin = -90
        self.ymax = 90

        self.times = times[0:-1]
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        # NB: assuming here constant time step
        self.time_step = self.times[-1] - self.times[-2]

        lon1 = lons[0:-1]
        lon2 = lons[1::]
        lat1 = lats[0:-1]
        lat2 = lats[1::]
        self.lon = (lon1+lon2)/2
        self.lat = (lat1+lat2)/2

        g = pyproj.Geod(ellps='WGS84')
        self.azimuth, backazimuth, self.dist = \
            g.inv(lon1, lat1, lon2, lat2, radians=False)
        self.speed = []
        for i, t in enumerate(self.dist):
            self.speed.append(self.dist[i] /
                              (times[i+1] - times[i]).total_seconds())
        self.speed = np.array(self.speed)
        self.x_sea_water_velocity = self.speed*np.sin(np.radians(self.azimuth))
        self.y_sea_water_velocity = self.speed*np.cos(np.radians(self.azimuth))

        # Subtract wind/stokes contribution
        if waves is not None:
            ts = waves.get_timeseries_at_position(lon=self.lon, lat=self.lat,
                    #start_time=self.start_time, end_time=self.end_time,
                    times=self.times,
                    variables=['sea_surface_wave_significant_height',
                               'sea_surface_wave_stokes_drift_x_velocity',
                               'sea_surface_wave_stokes_drift_y_velocity',
                               'sea_surface_wave_mean_period_from_variance_spectral_density_inverse_frequency_moment'])
            # NB: Stokes drift already rotated to lat-lon axes

            if z <= 0:
                print('Calculating Stokes drift at depth of %s m' % -z)
                from opendrift.models.physics_methods import stokes_drift_profile_breivik

                wave_peak_period = ts[
                    'sea_surface_wave_mean_period_from_variance_spectral_density_inverse_frequency_moment']
                    #'sea_surface_wave_period_at_variance_spectral_density_maximum']
                self.stokes_u, self.stokes_v, S = stokes_drift_profile_breivik(
                    ts['sea_surface_wave_stokes_drift_x_velocity'],
                    ts['sea_surface_wave_stokes_drift_y_velocity'],
                    ts['sea_surface_wave_significant_height'],
                    wave_peak_period, z)  # Should be mean period, not peak

                print('Subtracting Stokes drift from current')
                stokes_speed = np.sqrt(self.stokes_u*self.stokes_u+self.stokes_v*self.stokes_v)
                print('Current in range %s - %s' % (self.speed.min(), self.speed.max()))
                print('Stokesdrift in range %s - %s' % (stokes_speed.min(), stokes_speed.max()))
                print('Stokesdrift fraction in range %s - %s' % (
                        (stokes_speed/self.speed).min(), (stokes_speed/self.speed).max()))
                self.x_sea_water_velocity = self.x_sea_water_velocity - self.stokes_u
                self.y_sea_water_velocity = self.y_sea_water_velocity - self.stokes_v

        # Run constructor of parent Reader class
        super(Reader, self).__init__()


    def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None):

        requested_variables, time, x, y, z, outside = self.check_arguments(
            requested_variables, time, x, y, z)

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        variables = {}

        l = np.ones(len(x))
        variables['x_sea_water_velocity'] = \
            self.x_sea_water_velocity[indxTime]*l
        variables['y_sea_water_velocity'] = \
            self.y_sea_water_velocity[indxTime]*l

        return variables



"""
dataclass to store the compounds
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List

from ..common.measurement import MassFraction


@dataclass_to_json
@dataclass
class Compound:
    '''
        Some compounds that will be handled by this dataclass:
        - sulfur_mass_fraction: MassFraction = None
        - carbon_mass_fraction: MassFraction = None
        - hydrogen_mass_fraction: MassFraction = None
        - mercaptan_sulfur_mass_fraction: MassFraction = None
        - nitrogen_mass_fraction: MassFraction = None
        - ccr_percent: MassFraction = None  # conradson carbon residue
        - calcium_mass_fraction: MassFraction = None
        - hydrogen_sulfide_concentration: MassFraction = None
        - salt_content: MassFraction = None
        - paraffin_volume_fraction: MassFraction = None
        - naphthene_volume_fraction: MassFraction = None
        - aromatic_volume_fraction: MassFraction = None
    '''
    name: str = ""
    groups: list = field(default_factory=list)
    method: str = ""
    measurement: MassFraction = None
    comment: str = ""


class CompoundList(JSON_List):
    item_type = Compound



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Gaute Hope, MET Norway

import numpy as np
from datetime import datetime, timedelta
from netCDF4 import Dataset, MFDataset
import logging
logger = logging.getLogger(__name__)

from opendrift.readers.basereader import BaseReader, UnstructuredReader


class Reader(BaseReader, UnstructuredReader):
    """
    A reader for unstructured SHYFEM (irregularily gridded) `CF compliant
    <https://cfconventions.org/>`_ netCDF files.

    http://www.ismar.cnr.it/shyfem

    Args:
        :param filename: A single netCDF file, or a pattern of files. The
                         netCDF file can also be an URL to an OPeNDAP server.
        :type filename: string, requiered.

        :param name: Name of reader
        :type name: string, optional

    .. seealso::

        py:mod:`opendrift.readers.basereader.unstructured`.
    """

    variable_aliases = {
        'eastward_sea_water_velocity': 'x_sea_water_velocity',
        'northward_sea_water_velocity': 'y_sea_water_velocity',
        'sea_floor_depth_below_sea_surface': 'sea_floor_depth_below_sea_level'
    }

    dataset = None

    def __init__(self, filename=None, name=None):
        if filename is None:
            raise ValueError('Filename is missing')
        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        # xarray currently does not handle this type of grid:
        # https://github.com/pydata/xarray/issues/2233

        self.timer_start("open dataset")
        logger.info('Opening dataset: ' + filestr)
        if ('*' in filestr) or ('?' in filestr) or ('[' in filestr):
            logger.info('Opening files with MFDataset')
            self.dataset = MFDataset(filename)
        else:
            logger.info('Opening file with Dataset')
            self.dataset = Dataset(filename, 'r')

        self.proj4 = '+proj=lonlat'

        logger.info('Reading grid and coordinate variables..')

        self.x, self.y = self.dataset['longitude'][:], self.dataset[
            'latitude'][:]

        ref_time = datetime.fromisoformat(self.dataset['time'].units[14:33])
        
        self.times = np.array([
            ref_time + timedelta(seconds=d.item())
            for d in self.dataset['time'][:]
        ])
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        # time steps are not constant

        self.xmin = np.min(self.x)
        self.xmax = np.max(self.x)
        self.ymin = np.min(self.y)
        self.ymax = np.max(self.y)

        # levels are the depth of the bottom of each layer. re-assign to middle of layer
        # for nearest interpolation.
        self.z = -self.dataset['level'][:]
        self.z = np.insert(self.z, 0, [0.])
        self.z = self.z[:-1] + (np.diff(self.z) / 2)
        assert len(self.z) == len(self.dataset['level'][:])
        self.zmin, self.zmax = np.min(self.z), 0.
        assert (self.z <= 0).all()

        self.variable_mapping = {}
        for var_name in self.dataset.variables:
            # skipping coordinate variables
            if var_name in ['time', 'longitude', 'latitude', 'levels']:
                continue

            var = self.dataset[var_name]
            if 'standard_name' in var.ncattrs():
                std_name = var.getncattr('standard_name')
                std_name = self.variable_aliases.get(std_name, std_name)
                self.variable_mapping[std_name] = str(var_name)

        self.variables = list(self.variable_mapping.keys())

        # Run constructor of parent Reader class
        super().__init__()

        self.boundary = self._build_boundary_polygon_(self.x.compressed(),
                                                      self.y.compressed())

        self.timer_start("build index")
        logger.debug("building index of nodes..")
        self.nodes_idx = self._build_ckdtree_(self.x, self.y)
        self.timer_end("build index")

        self.timer_end("open dataset")

    def plot_mesh(self,corners=None):
        """
        Plot the grid mesh. Does not automatically show the figure.
        """
        import matplotlib.pyplot as plt
        plt.figure()
        plt.scatter(self.x, self.y, marker='x', color='blue', label='nodes')

        x, y = getattr(self.boundary, 'context').exterior.xy
        plt.plot(x, y, color='green', label='boundary')

        plt.legend()
        plt.title('Unstructured grid: %s\n%s' % (self.name, self.proj))
        plt.xlabel('lon [deg E]')
        plt.ylabel('lat [deg N]')

        if corners is not None:
            plt.xlim(corners[0],corners[1])
            plt.ylim(corners[2],corners[3])

    def get_variables(self,
                      requested_variables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        z = np.atleast_1d(z)
        #if len(z) == 1:
        #    z = z[0] * np.ones(x.shape)

        logger.debug("Requested variabels: %s, lengths: %d, %d, %d" %
                     (requested_variables, len(x), len(y), len(z)))

        requested_variables, time, x, y, z, _outside = \
            self.check_arguments(requested_variables, time, x, y, z)

        nearest_time, _time_before, _time_after, indx_nearest, _indx_before, _indx_after = self.nearest_time(
            time)

        logger.debug("Nearest time: %s" % nearest_time)

        variables = {}

        logger.debug("Interpolating node-variables..")

        nodes = self._nearest_node_(x, y)
        assert len(nodes) == len(x)

        for var in requested_variables:
            dvar = self.variable_mapping.get(var)
            logger.debug("Interpolating: %s (%s)" % (var, dvar))
            dvar = self.dataset[dvar]

            if len(dvar.shape) > 2:
                level_ind = self.__nearest_level__(z)

                # Reading the smallest block covering the actual data
                block = dvar[indx_nearest,
                             slice(nodes.min(),
                                   nodes.max() + 1),
                             slice(level_ind.min(),
                                   level_ind.max() + 1), ]

                # Picking the nearest value
                variables[var] = block[
                        nodes - nodes.min(),
                        level_ind - level_ind.min(),
                        ]
            elif len(dvar.shape) == 1:
                # Reading the smallest block covering the actual data
                block = dvar[slice(nodes.min(),
                                   nodes.max() + 1), ]

                # Picking the nearest value
                variables[var] = block[
                        nodes - nodes.min(),
                        ]
            else:
                logger.error('unknown dimensionality')

        return variables

    def __nearest_level__(self, z):
        """
        Find nearest index of z in levels.
        """
        return np.argmin(np.abs(self.z[:, None] - z), axis=0)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import numpy as np
import logging; logger = logging.getLogger(__name__)

from opendrift.models.oceandrift import OceanDrift, Lagrangian3DArray
#from opendrift.elements import LagrangianArray


# Defining the oil element properties
class PelagicEgg(Lagrangian3DArray):
    """Extending Lagrangian3DArray with specific properties for pelagic eggs
    """

    variables = Lagrangian3DArray.add_variables([
        ('diameter', {'dtype': np.float32,
                      'units': 'm',
                      'default': 0.0014}),  # for NEA Cod
        ('neutral_buoyancy_salinity', {'dtype': np.float32,
                                       'units': '[]',
                                       'default': 31.25}),  # for NEA Cod
        ('density', {'dtype': np.float32,
                     'units': 'kg/m^3',
                     'default': 1028.}),
        ('hatched', {'dtype': np.float32,
                     'units': '',
                     'default': 0.})])


class PelagicEggDrift(OceanDrift):
    """Buoyant particle trajectory model based on the OpenDrift framework.

        Developed at MET Norway

        Generic module for particles that are subject to vertical turbulent
        mixing with the possibility for positive or negative buoyancy

        Particles could be e.g. oil droplets, plankton, or sediments

        Under construction.
    """

    ElementType = PelagicEgg

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'sea_surface_wave_significant_height': {'fallback': 0},
        'sea_ice_area_fraction': {'fallback': 0},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'land_binary_mask': {'fallback': None},
        'sea_floor_depth_below_sea_level': {'fallback': 100},
        'ocean_vertical_diffusivity': {'fallback': 0.02, 'profiles': True},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_water_temperature': {'fallback': 10, 'profiles': True},
        'sea_water_salinity': {'fallback': 34, 'profiles': True},
        'surface_downward_x_stress': {'fallback': 0},
        'surface_downward_y_stress': {'fallback': 0},
        'turbulent_kinetic_energy': {'fallback': 0},
        'turbulent_generic_length_scale': {'fallback': 0},
        'upward_sea_water_velocity': {'fallback': 0},
      }

    # The depth range (in m) which profiles shall cover
    required_profiles_z_range = [-120, 0]

    # Default colors for plotting
    status_colors = {'initial': 'green', 'active': 'blue',
                     'hatched': 'red', 'eaten': 'yellow', 'died': 'magenta'}


    def __init__(self, *args, **kwargs):

        # Calling general constructor of parent class
        super(PelagicEggDrift, self).__init__(*args, **kwargs)

        # By default, eggs do not strand towards coastline
        self.set_config('general:coastline_action', 'previous')

        # Vertical mixing is enabled by default
        self.set_config('drift:vertical_mixing', True)

    def update_terminal_velocity(self, Tprofiles=None,
                                 Sprofiles=None, z_index=None):
        """Calculate terminal velocity for Pelagic Egg

        according to
        S. Sundby (1983): A one-dimensional model for the vertical
        distribution of pelagic fish eggs in the mixed layer
        Deep Sea Research (30) pp. 645-661

        Method copied from ibm.f90 module of LADIM:
        Vikebo, F., S. Sundby, B. Aadlandsvik and O. Otteraa (2007),
        Fish. Oceanogr. (16) pp. 216-228
        """
        g = 9.81  # ms-2

        # Pelagic Egg properties that determine buoyancy
        eggsize = self.elements.diameter  # 0.0014 for NEA Cod
        eggsalinity = self.elements.neutral_buoyancy_salinity
        # 31.25 for NEA Cod

        # prepare interpolation of temp, salt
        if not (Tprofiles is None and Sprofiles is None):
            if z_index is None:
                z_i = range(Tprofiles.shape[0])  # evtl. move out of loop
                # evtl. move out of loop
                z_index = interp1d(-self.environment_profiles['z'],
                                   z_i, bounds_error=False)
            zi = z_index(-self.elements.z)
            upper = np.maximum(np.floor(zi).astype(np.uint8), 0)
            lower = np.minimum(upper+1, Tprofiles.shape[0]-1)
            weight_upper = 1 - (zi - upper)

        # do interpolation of temp, salt if profiles were passed into
        # this function, if not, use reader by calling self.environment
        if Tprofiles is None:
            T0 = self.environment.sea_water_temperature
        else:
            T0 = Tprofiles[upper, range(Tprofiles.shape[1])] * \
                weight_upper + \
                Tprofiles[lower, range(Tprofiles.shape[1])] * \
                (1-weight_upper)
        if Sprofiles is None:
            S0 = self.environment.sea_water_salinity
        else:
            S0 = Sprofiles[upper, range(Sprofiles.shape[1])] * \
                weight_upper + \
                Sprofiles[lower, range(Sprofiles.shape[1])] * \
                (1-weight_upper)

        # The density difference bettwen a pelagic egg and the ambient water
        # is regulated by their salinity difference through the
        # equation of state for sea water.
        # The Egg has the same temperature as the ambient water and its
        # salinity is regulated by osmosis through the egg shell.
        DENSw = self.sea_water_density(T=T0, S=S0)
        DENSegg = self.sea_water_density(T=T0, S=eggsalinity)
        dr = DENSw-DENSegg  # density difference

        # water viscosity
        my_w = 0.001*(1.7915 - 0.0538*T0 + 0.007*(T0**(2.0)) - 0.0023*S0)
        # ~0.0014 kg m-1 s-1

        # terminal velocity for low Reynolds numbers
        W = (1.0/my_w)*(1.0/18.0)*g*eggsize**2 * dr

        # check if we are in a Reynolds regime where Re > 0.5
        highRe = np.where(W*1000*eggsize/my_w > 0.5)

        # Use empirical equations for terminal velocity in
        # high Reynolds numbers.
        # Empirical equations have length units in cm!
        my_w = 0.01854 * np.exp(-0.02783 * T0)  # in cm2/s
        d0 = (eggsize * 100) - 0.4 * \
            (9.0 * my_w**2 / (100 * g) * DENSw / dr)**(1.0 / 3.0)  # cm
        W2 = 19.0*d0*(0.001*dr)**(2.0/3.0)*(my_w*0.001*DENSw)**(-1.0/3.0)
        # cm/s
        W2 = W2/100.  # back to m/s

        W[highRe] = W2[highRe]
        self.elements.terminal_velocity = W

    def update(self):
        """Update positions and properties of buoyant particles."""

        # Turbulent Mixing
        self.update_terminal_velocity()
        self.vertical_mixing()

        # Horizontal advection
        self.advect_ocean_current()

        # Vertical advection
        if self.get_config('drift:vertical_advection') is True:
            self.vertical_advection()
